/*
 * Squeers Copyright (c) 2010, James Bailie.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *     * The name of James Bailie may not be used to endorse or promote
 * products derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Squeers is a multiplexing, kqueue-based, file server.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/event.h>
#include <sys/param.h>
#include <dirent.h>

#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>

#include <pwd.h>
#include <grp.h>
#include <time.h>

#include <fcntl.h>
#include <signal.h>
#include <syslog.h>
#include <stdarg.h>

#include <string.h>
#include <ctype.h>

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#define VERSION "1.3"
#define PIDFILE "/var/run/squeers.pid"

#define RECEIVE 0
#define RESPOND 1

/*
 * Size of the output queue used by kevent().
 */

#define QLEN 5000

/*
 * This is size of the IO buffers used in connection control blocks.
 * 32K is the default size of a socket send buffer on i386 FreeBSD 6.x.  We
 * try to move data in chunks of this size in an attempt to "fill the pipe"
 * of each connection.
 */

#define IOBUFSIZE 32768

/*
 * Size of input buffers to hold client requests.
 */

#define INBUFSIZE ( MAXPATHLEN + 1 )

/*
 * connection control block.
 */

struct ccb
{
   int source, sock, inactive, state;

   off_t count, written;

   char buffer[ IOBUFSIZE ], inbuf[ INBUFSIZE ], *inptr;
};

/*
 * KQueue Input and output queues.  We can generate a delete_timer event
 * for each item in the output queue.  We can also accept up to 500
 * connections when the listening socket has them ready to go.  Each added
 * connection adds three events to the input queue.
 */

#define INQ QLEN + 1500
struct kevent inqueue[ INQ ];
struct kevent outqueue[ QLEN ];

int in = 0, out = 0, closed = 0, timeout = 1, instances = 1, child = 0;

extern char *optarg;

int fd = -1, increment = 1024, logging = 0, testing = 0, backlog = 1000;

struct passwd *passwd;
struct group *group;

char *interface = "", *port = "1966", *grp = "nobody", *user = "nobody",
   *root = NULL;

struct stack *children = NULL;

/*
 * Stack which grows dynamically.
 */

struct stack
{
   int free, used;
   pid_t *values, *top;
};

void ev_set( int desc, short filter, u_short flags, struct ccb *item )
{
   struct kevent *kev;

   if ( in >= INQ )
      return;

   kev = &inqueue[ in++ ];

   kev->ident = desc;
   kev->filter = filter;
   kev->fflags = 0;
   kev->flags = flags;
   kev->udata = item;
}

void flip_events( struct ccb *item )
{
   ev_set( item->sock, EVFILT_READ, EV_DISABLE, item );
   ev_set( item->sock, EVFILT_WRITE, EV_ENABLE, item );
}

void delete_timer( struct ccb *item )
{
   struct kevent *kev;

   if ( in >= INQ )
      return;

   kev = &inqueue[ in++ ];

   kev->ident = item->sock;
   kev->filter = EVFILT_TIMER;
   kev->fflags = 0;
   kev->data = 0;
   kev->flags = EV_DELETE;
   kev->udata = item;
}

void set_timer( struct ccb *item )
{
   struct kevent *kev;

   if ( in >= INQ )
      return;

   kev = &inqueue[ in++ ];

   kev->ident = item->sock;
   kev->filter = EVFILT_TIMER;
   kev->fflags = 0;
   kev->data = timeout * 10000;
   kev->flags = EV_ADD;
   kev->udata = item;
}

void non_blocking( int desc )
{
   int flags, unblocked;

   if (( flags = fcntl( desc, F_GETFL, 0 )) < 0 )
   {
      syslog( LOG_ERR, "fcntl(): %m" );
      exit( 1 );
   }

   unblocked = flags & O_NONBLOCK;

   if ( ! unblocked && fcntl( desc, F_SETFL, flags | O_NONBLOCK ) < 0 )
   {
      syslog( LOG_ERR, "fcntl(): %m" );
      exit( 1 );
   }
}

void *memory( int size )
{
   void *ptr;

   if ( size == 0 )
      return NULL;

   if (( ptr = malloc( size )) == NULL )
   {
      if ( logging )
         syslog( LOG_WARNING, "malloc(): %m" );
      else
         fprintf( stderr, "squeers: malloc(): %s\n", strerror( errno ));
   }

   return ptr;
}

struct stack *make_stack()
{
   struct stack *a;

   if (( a = ( struct stack *)memory( sizeof( struct stack ))) == NULL )
      return NULL;

   if (( a->values = memory( sizeof( pid_t ) * increment )) == NULL )
      return NULL;

   a->free = increment;
   a->used = 0;
   a->top = a->values;

   return a;
}

int stack_push( struct stack *a, pid_t o )
{
   if ( a->free == 0 )
   {
      a->values = realloc( a->values, sizeof( pid_t ) *
                                      ( a->used + increment ) );

      if ( a->values == NULL )
      {
         if ( logging )
            syslog( LOG_WARNING, "realloc(): %m" );
         else
            fprintf( stderr, "squeers: realloc(): %s\n", strerror( errno ));

         return 1;
      }

      a->free = increment;
      a->top = &a->values[ a->used - 1 ];
   }

   if ( a->used )
      ++a->top;

   *a->top = o;
   --a->free;
   ++a->used;

   return 0;
}

pid_t stack_pop( struct stack *a )
{
   pid_t pid;

   if ( a->used )
   {
      pid = *a->top;
      --a->used;
      ++a->free;

      if ( a->used )
         --a->top;

      return pid;
   }

   return 0;
}

void signal_handler( int signo )
{
   while( children->used )
      kill( stack_pop( children ), SIGTERM );

   exit( 0 );
}

void set_options( int argc, char **argv )
{
   int i;

   while(( i = getopt( argc, argv, "xq:t:i:p:r:u:g:m:")) != -1 )
      switch( i )
      {
         case '?':
            exit( 1 );

         case 'g':
            grp = optarg;
            break;

         case 'i':
            interface = optarg;
            break;

         case 'm':
            instances = atoi( optarg );
            break;

         case 'p':
            port = optarg;
            break;

         case 'r':
            root = optarg;
            break;

         case 't':
            timeout = atoi( optarg );
            break;

         case 'u':
            user = optarg;
            break;

         case 'q':
            backlog = atoi( optarg );
            break;

         case 'x':
            testing = 1;
      }

   if ( root == NULL )
   {
      fputs( "squeers: server root directory not specified\n", stderr );
      exit( 1 );
   }

   if ( instances < 1 )
   {
      fprintf( stderr, "squeers: instances to run < 1: %d\n", instances );
      exit( 1 );
   }

   if ( timeout < 1 || timeout > 30 )
   {
      fprintf( stderr, "squeers: timeout value %d out of range 1-30\n", timeout );
      exit( 1 );
   }

   if (( passwd = getpwnam( user )) == NULL )
   {
      fprintf( stderr, "squeers: user \"%s\" does not exist\n", user );
      exit( 1 );
   }

   if (( group = getgrnam( grp )) == NULL )
   {
      fprintf( stderr, "squeers: group \"%s\" does not exist\n", grp );
      exit( 1 );
   }
}

void set_signals()
{
   int *iptr;
   int sigs[] = {
      SIGPIPE, SIGHUP, SIGQUIT, SIGUSR1, SIGUSR2, SIGALRM, SIGINT, SIGTSTP, -1
   };

   /*
    * Setting SIGCHLD to SIG_IGN prevents zombies from occurring.  We don't
    * need to wait() for our children.
    */

   signal( SIGCHLD, SIG_IGN );

   /*
    * We ignore everything but SIGTERM when ! testing.
    * Otherwise, we catch all the signals, so we may kill our children
    * before exiting.
    */

   signal( SIGTERM, signal_handler );

   for( iptr = sigs; *iptr > 0; ++iptr )
      signal( *iptr, ( testing ? signal_handler : SIG_IGN ));
}

void become_daemon()
{
   int file;
   char buffer[ 16 ];

   /*
    * Fork and let the parent die, continuing as child so we are not
    * a process group leader.  This is necessary for the call to setsid().
    */

   switch( fork() )
   {
      case -1:
         fprintf( stderr, "squeers: fork(): %s\n", strerror( errno ));
         exit( 1 );

      case 0:
         break;

      default:
         exit( 0 );
   }

   fclose( stdout );
   fclose( stderr );
   fclose( stdin );

   stdin = fopen( "/dev/null", "r" );
   stdout = fopen( "/dev/null", "w" );
   stderr = fopen( "/dev/null", "w" );

   if ( stdin == NULL || stdout == NULL || stderr == NULL )
   {
      syslog( LOG_ERR, "fopen(): %m" );
      exit( 1 );
   }

   if ( setsid() < 0 )
   {
      syslog( LOG_ERR, "setsid(): %m" );
      exit( 1 );
   }

   snprintf( buffer, sizeof( buffer ), "%d", getpid() );

   if (( file = open( PIDFILE, O_WRONLY | O_CREAT | O_TRUNC,
                      S_IWUSR | S_IRUSR | S_IRGRP )) < 0 )
      syslog( LOG_WARNING, "open( %s ): %m", PIDFILE );
   else
   {
      write( file, buffer, strlen( buffer ));
      close( file );
   }
}

void change_identity()
{
   if ( setgid( group->gr_gid ) < 0 )
   {
      syslog( LOG_ERR, "setgid(): %m" );
      exit( 1 );
   }

   if ( setuid( passwd->pw_uid ) < 0 )
   {
      syslog( LOG_ERR, "setuid(): %m" );
      exit( 1 );
   }
}

void start_listening()
{
   struct addrinfo hints, *res;
   struct accept_filter_arg afa;
   int result;

   bzero( &hints, sizeof( struct addrinfo ));
   hints.ai_flags = AI_PASSIVE;
   hints.ai_socktype = SOCK_STREAM;

   /*
    * If the user has not specified a listening interface, we use
    * the IPv6 wildcard.
    */

   if (( result = getaddrinfo( ( *interface ? interface : NULL ), port,
                               &hints, &res )))
   {
      syslog( LOG_ERR, "getaddrinfo(): %s", gai_strerror( result ));
      exit( 1 );
   }

   if ( res == NULL )
   {
      syslog( LOG_ERR, "getaddrinfo(): no interface found" );
      exit( 1 );
   }

   fd = socket( res->ai_family, res->ai_socktype, res->ai_protocol );

   if ( fd == -1 )
   {
      syslog( LOG_ERR, "socket(): %m" );
      exit( 1 );
   }

   result = 1;

   if ( setsockopt( fd, SOL_SOCKET, SO_REUSEADDR, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( SO_REUSEADDR ): %m" );

   if ( setsockopt( fd, SOL_SOCKET, SO_KEEPALIVE, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( SO_KEEPALIVE ): %m" );

   result = 0;

   /*
    * Note we are shutting off this option if we are binding to the IPv6
    * wildcard interface, so we may accept both IPv4 and IPv6 connections.
    */

   if ( ! *interface &&
        setsockopt( fd, IPPROTO_IPV6, IPV6_BINDV6ONLY, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( IPV6_BINDV6ONLY ): %m" );

   if ( bind( fd, res->ai_addr, res->ai_addrlen ) < 0 )
   {
      syslog( LOG_ERR, "bind(): %m" );
      exit( 1 );
   }

   if ( res != NULL )
      freeaddrinfo( res );

   if ( listen( fd, backlog ) < 0 )
   {
      syslog( LOG_ERR, "listen(): %m" );
      exit( 1 );
   }

   /*
    * Accept filter must be installed after socket is listening.
    */

   bzero( &afa, sizeof( afa ));
   strcpy( afa.af_name, "httpready" );

   if ( setsockopt( fd, SOL_SOCKET, SO_ACCEPTFILTER, &afa,
                    sizeof( afa )) < 0 )
      syslog( LOG_WARNING, "setsockopt( SO_ACCEPTFILTER ): %m" );
}

void add_conn( int new, char *addr )
{
   struct ccb *ptr;

   /*
    * Drop the connection if we have run out of memory.
    */

   if (( ptr = memory( sizeof( struct ccb ))) == NULL )
   {
      close( new );
      return;
   }

   *ptr->buffer = '\0';
   *ptr->inbuf = '\0';
   ptr->inptr = ptr->inbuf;

   ptr->count = ptr->written = 0;

   ptr->state = RECEIVE;
   ptr->source = -1;
   ptr->inactive = 0;
   ptr->sock = new;

   ev_set( ptr->sock, EVFILT_READ, EV_ADD, ptr );
   ev_set( ptr->sock, EVFILT_WRITE, EV_ADD | EV_DISABLE, ptr );
   set_timer( ptr );
}

void remove_conn( struct ccb *item )
{
   if ( item->source > 0 )
      close( item->source );

   closed = item->sock;
   close( item->sock );
   delete_timer( item );

   free( item );
}

void receive_request( struct ccb *item )
{
   struct stat st;
   off_t written = 0;
   int r;

   for( ; ; )
   {
      if ( !*item->inptr )
      {
         r = read( item->sock, item->inbuf, INBUFSIZE - 1 );

         if ( r <= 0 )
         {
            if ( r < 0 && ( errno == EWOULDBLOCK || errno == EAGAIN ))
            {
               item->inactive = 1;
               return;
            }
         }

         item->buffer[ r ] = '\0';
         item->inptr = item->inbuf;
      }

      /*
       * We don't want the line terminators.
       */

      if ( *item->inptr == 13 )
      {
         ++item->inptr;
         continue;
      }

      if ( *item->inptr == 10 )
         break;

      /*
       * New character overwrites previous string terminator.  New string
       * terminator added.  Notice that item->count does not count the
       * string terminator.
       */

      item->buffer[ item->count++ ] = *item->inptr++;

      if ( item->count >= IOBUFSIZE )
      {
         remove_conn( item );
         return;
      }

      item->buffer[ item->count ] = '\0';
   }

   flip_events( item );

   if ( stat( item->buffer, &st ) < 0 || ! S_ISREG( st.st_mode ))
   {
      remove_conn( item );
      return;
   }

   if (( item->source = open( item->buffer, O_RDONLY )) < 0 )
   {
      remove_conn( item );
      return;
   }

   item->state = RESPOND;
   item->count = st.st_size;

   if ( sendfile( item->source, item->sock, item->written,
                  MIN( item->count, IOBUFSIZE ), NULL,
                  &written, 0 ) < 0 )
   {
      if ( errno != EWOULDBLOCK && errno != EBUSY && errno != EAGAIN )
      {
         remove_conn( item );
         return;
      }
   }

   item->written += written;
   item->count -= written;

   if ( ! item->count )
      remove_conn( item );
}

int accept_connection()
{
   struct sockaddr_storage addr;
   char address[ 128 ];
   socklen_t len;
   int conn;

   len = sizeof( struct sockaddr_storage );
   conn = accept( fd, ( struct sockaddr *)&addr, &len );

   if ( conn < 0 )
   {
      if ( errno != EWOULDBLOCK && errno != EAGAIN )
         syslog( LOG_ERR, "accept(): %m" );

      return 1;
   }

   non_blocking( conn );
   add_conn( conn, address );
   return 0;
}

void remove_events( int n)
{
   while( ++n < out )
      if ( outqueue[ n ].ident == closed )
         outqueue[ n ].ident = -1;
}

/*
 * If a timer fires on a descriptor and that descriptor becomes readable
 * after the last invocation of kevent(), we could get the both events in
 * the queue, on the next invocation.  If the read event was earlier in the
 * queue, and in processing the read event, it was decided to close the
 * connection, its ccb would be freed with a call to remove_conn().  When
 * the timer event was encountered a second freeing of the ccb would occur.
 * If the events were placed in the queue in the opposite order, then
 * processing the timer event would call remove_conn(), and then the read
 * event handler would be dispatched with a pointer to the freed ccb.
 * Both situations will crash the server.

 * Our solution is to modify the output queue every time a connection is
 * dropped during the processing of a read, write, or timer event.  We make
 * sure any other events for the same descriptor further along in the queue
 * are ignored by setting the ident member of the appropriate struct kevent
 * to -1.

 * kevent() returns -1 when there is no room to create an error event in
 * the output queue.  We treat this case the same as an error event in the
 * output queue, and just move on.  Errors are generated from the
 * occasional failure to create a read or a write event on a connection
 * unexpectedly dropped by the client or server.  When kevent() does return
 * -1, the output queue may contain events, but we don't know how many, so
 * we can't service them.  If they are errors, we don't care.  If they are
 * IO events, they are level-triggered, so we will get those events again.
 * If they are timer events, they will be generated again, for timers are
 * periodic.  The worst that can happen is that a timeout will take longer
 * than otherwise to occur.  No events are lost.

 * Because we the listening socket can only appear once in the output queue,
 * the rate at which we accept new connections is inversely proportional to
 * the number of active connections, but we do try and accept up to 500 at
 * once, if they are ready to be accepted.
 */

void process_clients()
{
   int n, kq;
   struct ccb *item;

   if (( kq = kqueue()) < 0 )
   {
      syslog( LOG_ERR, "kqueue(): %m" );
      exit( 1 );
   }

   non_blocking( fd );
   ev_set( fd, EVFILT_READ, EV_ADD | EV_ENABLE, NULL );

   for( ; ; )
   {
      out = kevent( kq, inqueue, in, outqueue, QLEN, NULL );

      in = 0;

      if ( out <= 0 )
         continue;

      for( n = 0; n < out; ++n )
      {
         if ( outqueue[ n ].ident < 0 || outqueue[ n ].flags & EV_ERROR )
            continue;

         closed = 0;
         item = ( struct ccb *)outqueue[ n ].udata;

         if ( outqueue[ n ].filter == EVFILT_TIMER )
            remove_conn( item );
         else if ( outqueue[ n ].ident == fd )
         {
            int i;

            for( i = 0; i < 500; ++i )
               if ( accept_connection() )
                  break;
         }
         else if ( item->state == RECEIVE )
            receive_request( item );
         else
         {
            off_t written = 0;

            if ( sendfile( item->source, item->sock, item->written,
                           MIN( item->count, IOBUFSIZE ), NULL,
                           &written, 0 ) < 0 )
            {
               if ( errno != EWOULDBLOCK && errno != EBUSY && errno != EAGAIN )
               {
                  remove_conn( item );
                  continue;
               }
            }

            item->written += written;
            item->count -= written;

            if ( ! item->count )
               remove_conn( item );
         }

         if ( closed )
            remove_events( n );
      }
   }
}

int fork_children()
{
   pid_t pid;

   if (( children = make_stack()) == NULL )
      exit( 1 );

   while( --instances )
      switch(( pid = fork() ))
      {
         case -1:
            syslog( LOG_ERR, "fork(): %m" );
            /* fall through intended */

         case 0:
            return 1;

         default:
            if ( stack_push( children, pid ))
               exit( 1 );
      }

   return 0;
}

void change_directory()
{
   if ( chdir( root ) < 0 )
   {
      syslog( LOG_ERR, "chdir( %s ): %m", root );
      exit( 1 );
   }

   if ( chroot( root ) < 0 )
   {
      syslog( LOG_ERR, "chroot( %s ): %m", root );
      exit( 1 );
   }
}

int main( int argc, char **argv )
{
   set_options( argc, argv );
   set_signals();

   openlog( "squeers", LOG_PID, LOG_DAEMON );
   logging = 1;

   if ( !testing )
      become_daemon();

   start_listening();
   change_directory();

   change_identity();
   fork_children();

   process_clients();

   return 0;
}
